# Copyright (C) 2021 Taavi Väänänen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import base64
import ipaddress
import os
import pathlib
from typing import Dict

import hvac
import requests
from flask import current_app
from python_graphql_client import GraphqlClient

import preseedd

NETBOX_QUERY = """
query findServer($mac: String!) {
    vm_interface_list(filters: {mac_address: [$mac]}) {
        id
        virtual_machine {
            name
            config_context
            custom_fields
            site {
                custom_fields
            }
            device {
                site {
                   custom_fields
                }
            }
            primary_ip4 {
                address
                dns_name
            }
            primary_ip6 {
                address
                dns_name
            }
            platform {
                custom_fields
            }
        }
    }

    interface_list(filters: {mac_address: [$mac]}) {
        id
        device {
            name
            config_context
            custom_fields
            site {
                custom_fields
            }
            primary_ip4 {
                address
                dns_name
            }
            primary_ip6 {
                address
                dns_name
            }
            platform {
                custom_fields
            }
        }
    }
}
"""


class DeviceNotFoundError(Exception):
    """Raised when a device is not found for the given parameters."""

    pass


def get_kea_lease(ip: str):
    data = requests.post(
        current_app.config["KEA_CTRL_URL"],
        json={"service": ["dhcp4"], "command": "lease4-get-all"},
    )

    leases = [
        lease
        for lease in data.json()[0]["arguments"]["leases"]
        if lease["ip-address"] == ip
    ]

    if len(leases) == 1:
        return leases[0]

    raise DeviceNotFoundError(f"Failed to find a DHCP lease for IP {ip}")


def netbox_gql_client() -> GraphqlClient:
    token = current_app.config["NETBOX"]["API_TOKEN"]
    return GraphqlClient(
        endpoint=current_app.config["NETBOX"]["GQL_ENDPOINT"],
        headers={
            "User-Agent": f"preseedd/{preseedd.__version__}",
            "Authorization": f"Token {token}",
            # why is this needed???
            "Content-Type": "application/json",
        },
    )


def get_vm_data(mac: str):
    results = netbox_gql_client().execute(
        query=NETBOX_QUERY,
        operation_name="findServer",
        variables={"mac": mac.upper()},
    )

    if len(results["data"]["interface_list"]) == 1:
        return results["data"]["interface_list"][0]["device"]
    elif len(results["data"]["vm_interface_list"]) == 1:
        return results["data"]["vm_interface_list"][0]["virtual_machine"]
    else:
        raise DeviceNotFoundError(f"Did not find a device for MAC address {mac}")


def get_debian_repo_data(codename: str) -> Dict:
    config = current_app.config["DEBIAN_REPO"]
    key_file = pathlib.Path(config["KEY_FILE"]).read_bytes()
    key_data = base64.b64encode(key_file).decode("utf-8")

    return {
        "version": codename,
        "repository_url": config["BASE"],
        "repository_components": config["COMPONENTS"],
        "extra_packages": config.get("EXTRA_PACKAGES", []),
        "key_data": f"base64://{key_data}",
        "proxy": config.get("PROXY", ""),
    }


def locate_mac(ip: str) -> str:
    lease = get_kea_lease(ip)
    return lease["hw-address"]


def get_site(device: dict) -> dict:
    if device.get("site"):
        return device["site"]
    if device.get("device") and device["device"].get("site"):
        return device["device"]["site"]
    return {}


def get_vault_client() -> hvac.Client:
    token = os.environ.get("VAULT_TOKEN", None)
    if not token:
        token = pathlib.Path(current_app.config["VAULT"]["TOKEN_FILE"]).read_text()

    return hvac.Client(current_app.config["VAULT"]["ENDPOINT"], token=token)


def generate_root_password(hostname: str, site: str) -> str:
    vault_client = get_vault_client()
    vault_key = f"{site}/{hostname}"

    new_password = vault_client.read(
        f"sys/policies/password/{current_app.config['VAULT']['PASSWORD_POLICY']}/generate"
    )["data"]["password"]

    vault_client.secrets.kv.v2.create_or_update_secret(
        mount_point=current_app.config["VAULT"]["PREFIX"],
        path=vault_key,
        secret={"password": new_password},
    )

    return new_password


def get_all_host_data(mac: str, generate_password: bool = False) -> Dict:
    vm_data = get_vm_data(mac)

    hostname, domain = vm_data["primary_ip4"]["dns_name"].split(".", 1)

    ipv4 = ipaddress.IPv4Interface(vm_data["primary_ip4"]["address"])

    codename = vm_data["platform"]["custom_fields"]["debian_codename"]
    partition_scheme = vm_data["custom_fields"].get("disk_partition_layout", "custom")
    site = get_site(vm_data)

    data = {
        "nameservers": "10.20.100.53",
        "ntp_server": current_app.config.get("NTP_SERVER", "0.debian.pool.ntp.org"),
        "hostname": hostname,
        "domain": domain,
        "ipv4": {
            "address": str(ipv4.ip),
            "gateway": str(ipv4.network.network_address + 1),
            "netmask": str(ipv4.netmask),
        },
        "debian": get_debian_repo_data(codename),
        "partition_primary_drive": vm_data["config_context"].get(
            "partition_primary_drive"
        ),
        "partition_scheme": partition_scheme,
        "root_ssh_key": current_app.config["INITIAL_SSH_KEY"],
    }

    if generate_password:
        data["root_password"] = generate_root_password(
            hostname, site["custom_fields"]["puppet_site_id"]
        )

    return data
